package sxwaba.ui;

import waba.ui.*;
import waba.fx.*;
import waba.sys.Vm;
import waba.util.Vector;
import waba.util.IntVector;

/**
* @author ian reinhart geiser  <geiseri@sourcextreme.com>
* @version "$Id: BoxBase.java,v 1.3 2004/03/31 02:36:03 geiseri Exp $"
* @date 2003
* @class BoxBase
* @brief This is the baseclass for the Box Layouts. The box layouts allow one to create advanced layouts that will scale to screens of different resolutions.
*
*   This library is free software; you can redistribute it and/or
*   modify it under the terms of the GNU Lesser General Public
*   License as published by the Free Software Foundation; either
*   version 2 of the License, or (at your option) any later version.
*
*   This library is distributed in the hope that it will be useful,
*   but WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   Lesser General Public License for more details.
*
*   You should have received a copy of the GNU Lesser General Public
*   License along with this library; if not, write to the Free Software
*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* An example of how this class is used is shown below:
* @code
* BoxBase MainBox = new VBox();
* BoxBase ChildBox = new HBox();
* waba.ui.Label label = new waba.ui.Label("Test");
* waba.ui.Button button = new waba.ui.Button("Test");
* edit = new Edit("");
* edit.setMode(Edit.CURRENCY);
* edit.setValidChars("1234567890.");
* edit.setKeyboard(Edit.KBD_CALCULATOR);
* ChildBox.add(label,0);		// Default horizontal size
* ChildBox.add(edit,2);		// Stretch 2x all other horizontal controls
* MainBox.add(ChildBox,4);	// Stretch 4x all other vertical controls
* MainBox.add(button,0);		// Default vertical size
* MainBox.setRect(getClientRect());	// Applies the actual layout to everything here.
* @endcode
*/
public abstract class BoxBase extends Container
{
		protected Vector m_controls = null;
		protected IntVector m_stretch = null;

		protected int m_pad = 2;
		protected int m_spc = 2;
		protected int m_total = 0;
		protected int m_hSpc = m_spc/2;

		/**
		* Default constructor for the boxbase.  This is not usually called directly but used via the HBox and VBox objects.
		*/
		public BoxBase()
		{
			super();
			m_controls = new Vector();
			m_stretch = new IntVector();
		}

		/**
		* Sets the spacing between the controls in the layout.  The default value is 2 pixels.
		*/
		public void setGap( int spc )
		{
			m_spc = spc;
			m_hSpc = spc/2;
		}
		/**
		* Returns the spacing between controls in the layout.
		*/
		public int gap()
		{
			return m_spc;
		}
		/** 
		* Sets the padding arround the control inside of the layout.
		*/
		public void setPadding( int pad )
		{
			m_pad = pad;
		}
		/**
		* Returns the padding arround the contron inside of the layout.
		*/
		public int padding()
		{
			return m_pad;
		}
		/**
		* Adds a Control to the layout with a normal stretch value of 1.
		*/
		public void add( Control c )
		{
			add( c, 1 );
		}

		/**
		* Adds a Control to the layout with a specific stretch value.  The value of 0 means
		* the controls preferred size will be used and the control will not be scaled at all.
		*/
		public void add( Control c, int stretch)
		{
			//waba.sys.Vm.debug("Adding control " + c.toString() );
			super.add( c );
			m_controls.add( c );
			m_stretch.add( stretch );
			m_total += stretch;
		}

		/**
		* Adds a stretch value to a Control at offset x.
		*/
		public void addStretch( int x, int stretch )
		{
			int oldStretch = m_stretch.items[x];
			m_stretch.items[x] = stretch;
			m_total -= oldStretch;
			m_total += stretch;
		}
		/**
		*  A convenience method for addStretch(int x, int stretch) that takes an actual Control as the first argument.
		*/
		public void addStretch( Control c, int stretch )
		{
			int idx = m_controls.find( c );
			addStretch( idx, stretch );
		}
		/**
		* Adds an empty slot in the layout. This can be used to center or right/left justify
		* controls.
		* The example below creates something like this:
		*@image html layoutexample_spacers.png "Fig 1. Advanced Layout Example with Justifications."
		* Example:
		* @code
		* BoxBase MainBox = new VBox();
		* BoxBase CenterBox = new HBox();
		* BoxBase LeftBox = new HBox();
		* BoxBase RightBox = new HBox();
		*
		* CenterBox.addSpacer();	// Add spacer.
		* CenterBox.add(new waba.ui.Label("This is in the center"),0);	// Create the control
		* CenterBox.addSpacer();	// Add final spacer.
		*
		* LeftBox.add(new waba.ui.Label("This is on the left"),0);	// Create the control
		* LeftBox.addSpacer();	// Add final spacer.
		*
		* RightBox.addSpacer();	// Add spacer.
		* RightBox.add(new waba.ui.Label("This is on the right"),0);	// Create the control
		*
		* MainBox.add(CenterBox,0);
		* MainBox.add(LeftBox,0);
		* MainBox.add(RightBox,0);
		* MainBox.setRect(getClientRect());
		* @endcode
		*/
		public void addSpacer()
		{
			add( new Container(), 1);
		}
		
		/**
		* A convenience method that returns the stretch value of a Control in the layout.
		*/
		public int stretch(Control c)
		{
			return stretch( m_controls.find( c ));
		}
		/**
		* Returns the stretch value for a Control in the layout at index x.
		*/
		public int stretch(int x)
		{
			return m_stretch.items[x];
		}
		/**
		* A convenience method that removes a Control from the layout.
		*/
		public void remove( Control c )
		{
			int idx = m_controls.find( c );
			remove( idx );
		}
		/**
		* Removes a Control from the layout at index x.
		*/
		public void remove( int x )
		{

			super.remove( ( Control ) m_controls.items[x] );
			m_controls.del( x );
			m_total -= m_stretch.items[x];
			m_stretch.del( x );
		}

		/**
		* Returns the preferred width of the layout.
		*/
// 		public int getPreferredWidth()
// 		{
// 			return super.getPreferredWidth() + m_spc;
// 		}
		/**
		* Returns the preferred height of the layout.
		*/
// 		public int getPreferredHeight()
// 		{
// 			return super.getPreferredHeight() + m_spc;
// 		}
}

